/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: /n/olympus/opt/users/CVS/ns-1.1-master/tcp-nr-paced.cc,v 1.3 1997/11/05 16:03:46 wuchang Exp $ (LBL)";
#endif

//
// newreno-tcp: a revised reno TCP source, without sack
//

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"
#include "trace.h"
#include "random.h"
#include "node.h"
#include "scoreboard.h"

class NewRenoPacedTcpAgent : public TcpAgent {
 public:
	NewRenoPacedTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
	virtual void output(int seqno, int reason);
	virtual void reset();
	virtual void closecwnd(int how);
        virtual void opencwnd();
        virtual void quench(int how);    
	void send(int force, int reason, int maxburst = 0);

 protected:
	u_int dupwnd_;
	int changes_;		/* 0 for fixing unnecessary fast retransmits */
				/* 1 for additional code from Allman, */
				/*  to implement other algorithms from */
				/*  Hoe's paper */
	int ecn_fix_;           /* 0 for buggy opencwnd after quench */
				/* 1 for non-buggy opencwnd after quench */
	int counter_;
	int backoff_;		/* The amount to back off */
	int linear_;		/* Increase sublinearly or not (misnamed)*/
	double scale_;
	int bw_alg_;		/* Increase depending on bw */
	int bw_alg_plus_;	/* Extend BW based increase to cwnd > 1 */
	double bw_percent_;	/* Percentage increase on opencwnd */
	void partialnewack(Packet *pkt);
	void set_rbsnd_timer();
	double timer_interval_; /* Fine-grained timer-interval */
};

static class NewRenoPacedTcpMatcher : public Matcher {
public:
	NewRenoPacedTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno-paced") == 0)
			return (new NewRenoPacedTcpAgent());
		return (0);
	}
} matcher_renopaced;

void NewRenoPacedTcpAgent::reset()
{
	TcpAgent::reset();
        scale_ = 1;
	counter_ = 0;
	set_rbsnd_timer();

}

int NewRenoPacedTcpAgent::window()
{
        //
        // reno: inflate the window by dupwnd_
        //      dupwnd_ will be non-zero during fast recovery,
        //      at which time it contains the number of dup acks
        //
	int win = int(cwnd_) + dupwnd_;
        if (win > int(wnd_))
                win = int(wnd_);
        return (win);
}

NewRenoPacedTcpAgent::NewRenoPacedTcpAgent() : dupwnd_(0), changes_(0)
{
	link_int("ns_tcpnewreno_paced", "changes", &changes_, 0);
	link_int("ns_tcpnewreno_paced", "ecn-fix", &ecn_fix_, 0);
	link_int("ns_tcpnewreno_paced", "backoff", &backoff_, 0);
	link_int("ns_tcpnewreno_paced", "linear", &linear_, 0);
	link_int("ns_tcpnewreno_paced", "bw-alg", &bw_alg_, 0);
	link_int("ns_tcpnewreno_paced", "bw-alg-plus", &bw_alg_plus_, 0);
	link_real("ns_tcpnewreno_paced", "bw-percent", &bw_percent_,0);
        link_time("ns_tcpnewreno_paced","timer-interval",&timer_interval_, 0);
}

/* 
 * Process a packet that acks previously unacknowleges data, but
 * does not take us out of Fast Retransmit.
 */
void NewRenoPacedTcpAgent::partialnewack(Packet* pkt)
{
        newtimer(pkt);
#ifdef notyet
        if (pkt->seqno_ == stp->maxpkts && stp->maxpkts > 0)
                stp->endtime = (float) realtime();
#endif
        last_ack_ = pkt->seqno_;
        highest_ack_ = last_ack_;
        if (seqno_ < last_ack_ + 1)
                seqno_ = last_ack_ + 1;
        if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
                rtt_active_ = 0;
                t_backoff_ = 1;
        }
}

void NewRenoPacedTcpAgent::recv(Packet *pkt)
{
	static int acked = 0;
	int new_ssthresh;
	static double ack2, ack3, basertt;

        /* Use first packet to calculate the RTT  --contributed by Allman */

        if (++acked == 1)
                basertt = Scheduler::instance().clock() - firstsent_;

        /* Estimate ssthresh based on the calculated RTT and the estimated
           bandwidth (using ACKs 2 and 3).  */

        else if (acked == 2)
                ack2 = Scheduler::instance().clock();
        else if (acked == 3)
        {       
                ack3 = Scheduler::instance().clock();  
                new_ssthresh = int((basertt * (size_ / (ack3 - ack2)))
                                    / size_);
                if (changes_ > 0 && new_ssthresh < ssthresh_)
                    ssthresh_ = new_ssthresh;
        }

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN) {
		quench(1);
		if (trace_) {
			sprintf(trace_->buffer(), "t %g RECV ECN\n",
				Scheduler::instance().clock());
			trace_->dump();
		}
	} else
		if (trace_) {
			sprintf(trace_->buffer(), "t %g RECV NOECN\n",
				Scheduler::instance().clock());
			trace_->dump();
		}
	if (trace_) {
		sprintf(trace_->buffer(), "t %g seqno %d lastack %d recover %d\n",
			Scheduler::instance().clock(), pkt->seqno_, last_ack_, recover_);
		trace_->dump();
	}
	if (pkt->seqno_ > last_ack_) {
	    if (pkt->seqno_ >= recover_) {
		dupwnd_ = 0;
		newack(pkt);
		if (!(pkt->flags_ & PF_ECN) || !ecn_fix_)
                	opencwnd();
	    }
	    else {
		/* received new ack for a packet sent during Fast
		 *  Recovery, but sender stays in Fast Recovery */
		dupwnd_ = 0;
		partialnewack(pkt);
		output(last_ack_ + 1, 0);
	    }
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				closecwnd(1);
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
			dupwnd_ = NUMDUPACKS;
		} else if (dupacks_ > NUMDUPACKS) 
			++dupwnd_;
                        /* For every two duplicate ACKs we receive (in the
                         * "fast retransmit phase"), send one entirely new
                         * data packet "to keep the flywheel going".  --Allman 
			 */
			 if (changes_ > 0 && (dupacks_ % 2) == 1)
		   		output (seqno_++,0);

	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0) 
                /*
                 * Maxburst is really only needed for the first
                 *   window of data on exiting Fast Recovery.
                 */
		send(0, 0, maxburst_);
	else if (dupacks_ > NUMDUPACKS - 1 && changes_==0)
		send(0, 0, 2);
}

/*
 * Called from DataSource
 */
void NewRenoPacedTcpAgent::output(int seqno, int reason)
{
	Packet* p = allocpkt(seqno);
	double now = Scheduler::instance().clock();

	p->bd_.tcp_.ts_ = now;
	p->flags_ |= reason;
	node_->transmit(p);
	if (seqno > maxseq_) {
		maxseq_ = seqno;
		if (!rtt_active_) {
			rtt_active_ = 1;
			if (seqno > rtt_seq_)
				rtt_seq_ = seqno;
		}
	}
	if (!pending_[TCP_TIMER_RTX])
		/* No timer pending.  Schedule one. */
		set_rtx_timer();
}

void NewRenoPacedTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
	}

	/* retransmit timer */
	if (tno == TCP_TIMER_RTX) {
		timeouts_++;
		recover_ = maxseq_;
		recover_cause_ = 2;
		closecwnd(0);
		reset_rtx_timer(0);
		send(0, PF_TIMEOUT);
	} else if (tno == TCP_TIMER_RBSND) {
		if (trace_) {
			sprintf(trace_->buffer(), "t %g \n",
				Scheduler::instance().clock());
			trace_->dump();
		}
                	counter_--;
                	if (counter_ <= 0) {
                    		counter_ = 0;
                    		send(0, PF_RBSND);
                	}
			if (pending_[TCP_TIMER_RBSND]) {
		  		fprintf(stderr,"rbsnd timer should have been cancelled\n");
		  		abort();
			}
		set_rbsnd_timer();
	}

}

/*
 * Try to send as much data as the window will allow.  The link layer will
 * do the buffering; we ask the application layer for the size of the packets.
 */
void NewRenoPacedTcpAgent::send(int force, int reason, int maxburst)
{
        int win = window();
        int npackets = 0;
        double time_bw_sends;

        time_bw_sends = (double) (scale_ * t_rtt_ * tcp_tick_)/cwnd_;
	if (trace_) {
		sprintf(trace_->buffer(), "t %g t_bw_s %g t_int %g\n",
			Scheduler::instance().clock(), time_bw_sends, 
			timer_interval_);
		trace_->dump();
	}
        if (time_bw_sends < timer_interval_)
            while (seqno_ <= highest_ack_ + win && seqno_ < curseq_) {
                if ( npackets < 2 ) {
			output(seqno_++, reason);
			npackets++;
                        // counter_ = 100000;
		} 
		else {
			return;
		}
            }
        else {
		if (trace_) {
			sprintf(trace_->buffer(), "t %g count %d\n",
				Scheduler::instance().clock(), counter_ );
			trace_->dump();
		}
            if (counter_ == 0 && seqno_ <= highest_ack_ + win && seqno_ < curseq_) {
                output(seqno_++,reason);
                counter_ = int(time_bw_sends/timer_interval_) + 1;
            }
        }
}
/*
 * open up the congestion window
 */
void NewRenoPacedTcpAgent::opencwnd()
{
	double eff_wnd_;
	// SubTCP code
	//   linear_ ==> do sublinear increases
	//   bw_alg_ ==> do sublinear increases based on bandwidth
    	if ( (cwnd_ == 1) && (scale_ > 1)) {
		if (linear_) {
			if (bw_alg_) {
				eff_wnd_ = (1+bw_percent_)/scale_;
				if (eff_wnd_ > 1) {
					cwnd_ = eff_wnd_;
					scale_ = 1;	
				}
				else {
					cwnd_ = 1;
					scale_ = 1/eff_wnd_;
				}
			}
			else
				scale_ -= (double) 1/linear_;
		}	
		else
			scale_ = scale_ / 2;
		if (scale_ < 1) scale_ = 1;
		return;
        }
	scale_ = 1;
	if (cwnd_ < ssthresh_) {
		/* slow-start (exponential) */
		cwnd_ += 1;
	} else {
		/* linear */
		double f;
		switch (wnd_option_) {
		case 0:
			if (++count_ >= cwnd_) {
				count_ = 0;
				++cwnd_;
			}
			break;

		case 1:
		/* This is the standard algorithm. */
			if (bw_alg_plus_)
				cwnd_ += bw_percent_;
			else
				cwnd_ += 1 / cwnd_;
			break;

		default:
#ifdef notdef
			/*XXX*/
			error("illegal window option %d", wnd_option_);
#endif
			abort();
		}
	}
	// if maxcwnd_ is set (nonzero), make it the cwnd limit
	if (maxcwnd_ && (int(cwnd_) > maxcwnd_))
		cwnd_ = maxcwnd_;

	return;

}

/*
 * close down the congestion window
 */
void NewRenoPacedTcpAgent::closecwnd(int how)
{
	switch (how) {
	case 0:
		/* timeouts, Tahoe dup acks */
		ssthresh_ = int( window() / 2 );
		/* Old code: ssthresh_ = int(cwnd_ / 2); */
		cwnd_ = int(wnd_init_);
		break;

	case 1:
		/* Reno dup acks, or after a recent congestion indication. */
		if ( cwnd_ == 1 ) {
                    scale_ = scale_ * backoff_;
		    if (scale_ > 64) scale_ = 64;
                    counter_ = int(scale_*t_rtt_*tcp_tick_/timer_interval_) + 1;
		    if (trace_) {
			sprintf(trace_->buffer(), "t %g ECN with cwnd 1, scale: %f counter_ %d\n", Scheduler::instance().clock(), scale_, counter_);
			trace_->dump();
		    }
		}
                else {
                    cwnd_ = window()/2;
                    scale_ = 1;
                }
		ssthresh_ = int(cwnd_);
		if (ssthresh_ < 2)
		  ssthresh_ = 2;
		break;

	case 2:
		/* Tahoe dup acks  		*
		 * after a recent congestion indication */
		cwnd_ = wnd_init_;
		break;

	default:
		abort();
	}
	fcnt_ = 0.;
	count_ = 0;
}

void NewRenoPacedTcpAgent::set_rbsnd_timer()
{
	if (pending_[TCP_TIMER_RBSND]) {
		cancel(TCP_TIMER_RBSND);
	}
	sched(timer_interval_,TCP_TIMER_RBSND);
}

/*
 * Respond either to a source quench or to a congestion indication bit.
 * This is done at most once a roundtrip time;  after a source quench,
 * another one will not be done until the last packet transmitted before
 * the previous source quench has been ACKed.
 */
void NewRenoPacedTcpAgent::quench(int how)
{
	if (highest_ack_ >= recover_) {
		recover_ = seqno_ - 1;
		recover_cause_ = 3;
		closecwnd(how);
		if (trace_) {
			double now = Scheduler::instance().clock();
			sprintf(trace_->buffer(),
				"%g pkt %d\n", now, seqno_ - 1);
			trace_->dump();
		}
	}
}
